# -*- coding: utf-8 -*-

from . cdiac import CDIACLoader
from . clio_infra import ClioInfraLoader
from . igme import IGMELoader
from . ihme import IHMELoader
from . ilo import ILOLoader
from . oecd import OECDLoader
from . worldbank import WorldBankLoader